# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor Order Files
"""

import decimal

from rattail.exceptions import RattailError
from rattail.excel import ExcelReaderXLSX


class OrderParserNotFound(RattailError):
    """
    Exception raised when an order file parser is required, but cannot
    be located.
    """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return f"Vendor order parser not found for key: {self.key}"


class OrderParser:
    """
    Base class for all vendor order parsers.
    """
    vendor_key = None

    def __init__(self, config):
        self.config = config
        self.app = self.config.get_app()

    @property
    def key(self):
        """
        Key for the parser.  Must be unique among all order parsers.
        """
        raise NotImplementedError(f"Order parser has no key: {repr(self)}")

    @property
    def title(self):
        """
        Human-friendly title for the parser.
        """
        raise NotImplementedError(f"Order parser has no title: {self.key}")

    def get_vendor(self, session):
        """
        Fetch the :class:`~rattail.db.model.vendors.Vendor` record
        which is associated with the current parser, if any.
        """
        if self.vendor_key:
            return self.app.get_vendor_handler().get_vendor(session, self.vendor_key)

    def parse_order_date(self, path):
        """
        Parse the order date from the order file.
        """

    def parse_order_number(self, path):
        """
        Parse the order number from the order file.
        """

    def parse_order_items(self, path, progress=None):
        """
        Parse all data items (rows) from the order file.
        """
        raise NotImplementedError

    def make_order_item(self, **kwargs):
        """
        Make and return a
        :class:`~rattail.db.model.purchase.PurchaseItem` instance.
        """
        model = self.app.model
        return model.PurchaseItem(**kwargs)


class ExcelOrderParser(OrderParser):
    """
    Base class for Excel vendor order parsers.
    """

    def get_excel_reader(self, path):
        """
        Return an :class:`~rattail.excel.ExcelReaderXLSX` instance for
        the given path.
        """
        if not hasattr(self, 'excel_reader'):
            kwargs = self.get_excel_reader_kwargs()
            self.excel_reader = ExcelReaderXLSX(path, **kwargs)
        return self.excel_reader

    def get_excel_reader_kwargs(self, **kwargs):
        """
        Should return kwargs for the Excel reader factory.
        """
        return kwargs

    def decimal(self, value, scale=2):
        """
        Convert a value to a decimal.
        """
        if value is None:
            return

        # no reason to convert integers, really
        if isinstance(value, (int, decimal.Decimal)):
            return value

        # float becomes rounded decimal
        if isinstance(value, float):
            return decimal.Decimal(f"{{:0.{scale}f}}".format(value))

        # string becomes decimal
        value = value.strip()
        if value:
            return decimal.Decimal(value)


class DefaultOrderParser(ExcelOrderParser):
    """
    Default order parser for Excel files.

    .. autoattribute:: key

    .. autoattribute:: title
    """
    key = 'default'
    title = "Default Excel Parser"

    # TODO: needs sane default parser logic



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Stuff
"""

import warnings

try:
    from .sess import Session, SessionBase as TrueSessionBase

except ImportError: # pragma: no cover
    Session = None
    SessionBase = None

else:

    class SessionBase(TrueSessionBase):

        def __init__(self, *args, **kwargs):
            warnings.warn("SessionBase should not be imported from rattail.db; "
                          "please import it from rattail.db.sess instead",
                          DeprecationWarning, stacklevel=2)
            super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-

from unittest import TestCase

from rattail.config import RattailConfig

try:
    import sqlalchemy as sa
    from rattail.autocomplete import customers as mod
    from rattail.db import Session
except ImportError:
    pass
else:

    class TestCustomerAutocompleter(TestCase):

        def setUp(self):
            self.config = RattailConfig()
            self.app = self.config.get_app()
            self.autocompleter = self.make_autocompleter()

        def make_autocompleter(self):
            return mod.CustomerAutocompleter(self.config)

        def test_autocomplete(self):
            engine = sa.create_engine('sqlite://')
            model = self.app.model
            model.Base.metadata.create_all(bind=engine)
            session = Session(bind=engine)
            enum = self.config.get_enum()

            # first create some customers
            alice = model.Customer(name='Alice Chalmers')
            session.add(alice)
            bob = model.Customer(name='Bob Loblaw')
            session.add(bob)
            charlie = model.Customer(name='Charlie Chaplin')
            session.add(charlie)

            # searching for nothing yields no results
            result = self.autocompleter.autocomplete(session, '')
            self.assertEqual(len(result), 0)

            # search for 'l' yields all 3 customers
            result = self.autocompleter.autocomplete(session, 'l')
            self.assertEqual(len(result), 3)

            # search for 'cha' yields just 2 customers
            result = self.autocompleter.autocomplete(session, 'cha')
            self.assertEqual(len(result), 2)
            uuids = [info['value'] for info in result]
            self.assertIn(alice.uuid, uuids)
            self.assertIn(charlie.uuid, uuids)



# -*- coding: utf-8; -*-
"""grow equity payment amount

Revision ID: b9d58e803208
Revises: f6e95c74d8db
Create Date: 2024-08-13 10:59:47.778670

"""

# revision identifiers, used by Alembic.
revision = 'b9d58e803208'
down_revision = 'f6e95c74d8db'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member_equity_payment
    op.alter_column('member_equity_payment', 'amount',
               existing_type=sa.NUMERIC(precision=7, scale=2),
               type_=sa.Numeric(precision=10, scale=2),
               existing_nullable=False)
    op.alter_column('member_equity_payment_version', 'amount',
               existing_type=sa.NUMERIC(precision=7, scale=2),
               type_=sa.Numeric(precision=10, scale=2),
               existing_nullable=True)


def downgrade():

    # member_equity_payment
    op.alter_column('member_equity_payment_version', 'amount',
               existing_type=sa.Numeric(precision=10, scale=2),
               type_=sa.NUMERIC(precision=7, scale=2),
               existing_nullable=True)
    op.alter_column('member_equity_payment', 'amount',
               existing_type=sa.Numeric(precision=10, scale=2),
               type_=sa.NUMERIC(precision=7, scale=2),
               existing_nullable=False)



# -*- coding: utf-8; -*-

from unittest import TestCase

import pytest

from rattail import db as mod


class TestModule(TestCase):

    def test_basic(self):
        try:
            from sqlalchemy import orm
        except ImportError:
            pytest.skip("test is not relevant without sqlalchemy")

        # Session is legitimate
        self.assertTrue(isinstance(mod.Session, orm.sessionmaker))

        # SessionBase warns of deprecation
        # (we don't test for that, just here for the coverage)
        session = mod.SessionBase()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2024 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Sessions
"""

import logging

import sqlalchemy
from sqlalchemy import orm


log = logging.getLogger(__name__)


class SessionBase(orm.Session):
    """
    Custom SQLAlchemy session base class, which adds some
    convenience methods related to the SQLAlchemy-Continuum
    integration.

    You should not instantiate this class directly; instead just
    use :class:`Session`.

    :param continuum_user: Optional user for Continuum versioning
       authorship.  If specified, the value is passed to
       :meth:`set_continuum_user()`.
    """

    def __init__(
            self,
            rattail_config=None,
            rattail_record_changes=None,
            continuum_user=None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.rattail_config = rattail_config

        # maybe record changes
        if rattail_record_changes is None:
            rattail_record_changes = getattr(self.bind, 'rattail_record_changes', False)
        if rattail_record_changes:
            from rattail.db.changes import record_changes
            record_changes(self, config=self.rattail_config)
        else:
            self.rattail_record_changes = False

        if continuum_user is None:
            self.continuum_user = None
        else:
            self.set_continuum_user(continuum_user)

        # maybe log the current db pool status
        if getattr(self.bind, 'rattail_log_pool_status', False):
            log.debug(self.bind.pool.status())

    def set_continuum_user(self, user_info):
        """
        Set the effective Continuum user for the session.

        :param user_info: May be a
          :class:`~rattail.db.model.users.User` instance, or the
          ``uuid`` or ``username`` for one.
        """
        if self.rattail_config:
            app = self.rattail_config.get_app()
            model = app.model
        else:
            from rattail.db import model

        if isinstance(user_info, model.User):
            user = self.merge(user_info)
        else:
            user = self.get(model.User, user_info)
            if not user:
                try:
                    user = self.query(model.User).filter_by(username=user_info).one()
                except orm.exc.NoResultFound:
                    user = None
        self.continuum_user = user


Session = orm.sessionmaker(class_=SessionBase, rattail_config=None, expire_on_commit=False)



# -*- coding: utf-8; -*-

from wuttjamaican.testing import FileTestCase
from rattail.config import RattailConfig


class DataTestCase(FileTestCase):
    """
    Base class for test suites requiring a full (typical) database.
    """

    def setUp(self):
        self.setup_db()

    def setup_db(self):
        self.setup_files()
        self.config = self.make_config(defaults={
            'rattail.db.default.url': 'sqlite://',
        })
        self.app = self.config.get_app()

        # init db
        model = self.app.model
        model.Base.metadata.create_all(bind=self.config.appdb_engine)
        self.session = self.app.make_session()

    def tearDown(self):
        self.teardown_db()

    def teardown_db(self):
        self.teardown_files()
        self.session.close()

    def make_config(self, **kwargs):
        return RattailConfig(**kwargs)



# -*- coding: utf-8; -*-

from rattail.trainwreck import config as mod
from rattail.testing import DataTestCase

try:
    import sqlalchemy
except ImportError:
    pass
else:

    class TestTrainwreckConfig(DataTestCase):

        def test_configure(self):
            self.assertFalse(hasattr(self.config, 'trainwreck_engines'))
            ext = mod.TrainwreckConfig()
            ext.configure(self.config)
            self.assertEqual(self.config.trainwreck_engines, {})



# -*- coding: utf-8; -*-
"""add PurchaseBatch.order_file

Revision ID: 6f0659414074
Revises: b9d58e803208
Create Date: 2024-10-21 14:53:02.779348

"""

# revision identifiers, used by Alembic.
revision = '6f0659414074'
down_revision = 'b9d58e803208'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_purchase
    op.add_column('batch_purchase', sa.Column('order_file', sa.String(length=255), nullable=True))
    op.add_column('batch_purchase', sa.Column('order_parser_key', sa.String(length=100), nullable=True))


def downgrade():

    # batch_purchase
    op.drop_column('batch_purchase', 'order_parser_key')
    op.drop_column('batch_purchase', 'order_file')



# -*- coding: utf-8; -*-

from unittest.mock import patch

from rattail import db
from rattail.testing import DataTestCase

try:
    import sqlalchemy as sa
    from rattail.db import sess as mod
except ImportError:
    pass
else:

    class TestSession(DataTestCase):

        def setUp(self):
            self.setup_db()

            model = self.app.model
            self.user = model.User(username='barney')
            self.session.add(self.user)
            self.session.commit()

        def test_init_rattail_config(self):
            mod.Session.configure(rattail_config=None)
            session = mod.Session()
            self.assertIsNone(session.rattail_config)
            session.close()

            config = object()
            session = mod.Session(rattail_config=config)
            self.assertIs(session.rattail_config, config)
            session.close()

        def test_init_record_changes(self):
            if hasattr(mod.Session, 'kw'):
                self.assertIsNone(mod.Session.kw.get('rattail_record_changes'))

            session = mod.Session()
            self.assertFalse(session.rattail_record_changes)
            session.close()

            session = mod.Session(rattail_record_changes=True)
            self.assertTrue(session.rattail_record_changes)
            session.close()

            engine = sa.create_engine('sqlite://')
            engine.rattail_record_changes = True
            session = mod.Session(bind=engine)
            self.assertTrue(session.rattail_record_changes)
            session.close()

        def test_continuum_user_param(self):

            # null by default
            session = mod.Session()
            self.assertIsNone(session.continuum_user)

            # can pass user
            session = mod.Session(continuum_user=self.user)
            self.assertEqual(session.continuum_user.uuid, self.user.uuid)

            # can pass username
            session = mod.Session(continuum_user='barney')
            self.assertEqual(session.continuum_user.uuid, self.user.uuid)

        def test_log_pool_status(self):

            # sanity/coverage check
            self.config.appdb_engine.rattail_log_pool_status = True
            session = mod.Session()

        def test_set_continuum_user(self):
            session = mod.Session()

            # null by default
            self.assertIsNone(session.continuum_user)

            # can pass user
            session.set_continuum_user(self.user)
            self.assertEqual(session.continuum_user.uuid, self.user.uuid)

            # can pass username
            session.set_continuum_user('barney')
            self.assertEqual(session.continuum_user.uuid, self.user.uuid)

            # null if bad username
            session.set_continuum_user('doesnotexist')
            self.assertIsNone(session.continuum_user)

            # still works if session has no config
            # TODO: not sure why session actually has a config here?
            # (guessing it is an artifact of running other tests?)
            self.assertIsNotNone(session.rattail_config)
            with patch.object(session, 'rattail_config', new=None):
                self.assertIsNone(session.rattail_config)
                session.set_continuum_user('barney')
            self.assertEqual(session.continuum_user.uuid, self.user.uuid)
